import errno
import traceback

from mercurial import revlog
from mercurial import node
from mercurial import context
from mercurial import util as hgutil
from mercurial import error

import compathacks
import svnexternals
import util


class MissingPlainTextError(Exception):
    """Exception raised when the repo lacks a source file required for replaying
    a txdelta.
    """

class ReplayException(Exception):
    """Exception raised when you try and commit but the replay encountered an
    exception.
    """

def updateexternals(ui, meta, current):
    # TODO fix and re-enable externals for single-directory clones
    if not current.externals or meta.layout == 'single':
        return

    # accumulate externals records for all branches
    revnum = current.rev.revnum
    branches = {}
    for path, entry in current.externals.iteritems():
        if not meta.is_path_valid(path):
            continue

        p, b, bp = meta.split_branch_path(path)
        if bp not in branches:
            parent = meta.get_parent_revision(revnum, b)
            pctx = meta.repo[parent]
            branches[bp] = (svnexternals.parse(ui, pctx), pctx)
        branches[bp][0][p] = entry

    # register externals file changes
    for bp, (external, pctx) in branches.iteritems():
        if bp and bp[-1] != '/':
            bp += '/'
        updates = svnexternals.getchanges(ui, meta.repo, pctx, external)
        for fn, data in updates.iteritems():
            path = (bp and bp + fn) or fn
            if data is not None:
                current.set(path, data, False, False)
            else:
                current.delete(path)

def convert_rev(ui, meta, svn, r, tbdelta, firstrun):
    try:
        return _convert_rev(ui, meta, svn, r, tbdelta, firstrun)
    finally:
        meta.editor.current.close()

def _convert_rev(ui, meta, svn, r, tbdelta, firstrun):

    editor = meta.editor
    editor.current.clear()
    editor.current.rev = r
    editor.setsvn(svn)

    if firstrun and meta.firstpulled <= 0:
        # We know nothing about this project, so fetch everything before
        # trying to apply deltas.
        ui.debug('replay: fetching full revision\n')
        svn.get_revision(r.revnum, editor)
    else:
        svn.get_replay(r.revnum, editor, meta.firstpulled)
    editor.close()

    current = editor.current

    updateexternals(ui, meta, current)

    if current.exception is not None:  # pragma: no cover
        traceback.print_exception(*current.exception)
        raise ReplayException()

    files_to_commit = current.files()
    branch_batches = {}
    rev = current.rev
    date = meta.fixdate(rev.date)
    mesg = meta.getmessage(rev)

    # build up the branches that have files on them
    failoninvalid = ui.configbool('hgsubversion',
            'failoninvalidreplayfile', False)
    for f in files_to_commit:
        if not meta.is_path_valid(f):
            if failoninvalid:
                raise hgutil.Abort('file %s should not be in commit list' % f)
            continue
        p, b = meta.split_branch_path(f)[:2]
        if b not in branch_batches:
            branch_batches[b] = []
        if p:
            branch_batches[b].append((p, f))

    closebranches = {}
    for branch in tbdelta['branches'][1]:
        branchedits = meta.revmap.branchedits(branch, rev)
        if len(branchedits) < 1:
            # can't close a branch that never existed
            continue
        ha = branchedits[0][1]
        closebranches[branch] = ha

    extraempty = (set(tbdelta['branches'][0]) -
                  (set(current.emptybranches) | set(branch_batches.keys())))
    current.emptybranches.update([(x, False) for x in extraempty])

    guessmerge = [word for word in meta.guessmerge
                  if word in mesg.lower()]

    # 1. handle normal commits
    closedrevs = closebranches.values()
    for branch, files in branch_batches.iteritems():

        if branch in current.emptybranches and files:
            del current.emptybranches[branch]

        if meta.skipbranch(branch):
            # make sure we also get rid of it from emptybranches
            if branch in current.emptybranches:
                del current.emptybranches[branch]
            continue

        files = dict(files)
        parents = [meta.get_parent_revision(rev.revnum, branch), revlog.nullid]

        hbranch = (-1, '')
        if current.mergeinfo:
            # Notes about computing the merge:
            # 1) we look for the top-most path (i.e. the branch name)
            #    which are the keys of the mergeinfo dict
            # 2) and see which has the highest rev (the are returned
            #    sorted lowest to high from the server)
            # 3) check to see if that branch is actually in the repo
            for bm, val in current.mergeinfo.iteritems():
                if val and meta.split_branch_path(bm)[1] in meta.branches:
                    # the value has the form:
                    # '/branches/foo:105,201-205\n/branches/bar:144,190-194'
                    for line in val.splitlines():
                        # get the branch name and comma-delimited list of
                        # revisions
                        tmp = line.split(':', 1)
                        # only keep the highest (last) one
                        tmp[1] = tmp[1].rsplit(',', 1)[-1]
                        # convert it to an int, possibly splitting a range
                        # (e.g. 201-205)
                        revrange = tmp[1].split('-', 1)
                        tmp[1] = int(revrange[-1])
                        if tmp[1] >= hbranch[0]:
                            # find the highest rev on the second parent branch
                            name = meta.split_branch_path(tmp[0])[1]
                            for lr in range(tmp[1], int(revrange[0]), -1):
                                if (lr, name) in meta.revmap:
                                    tmp[1] = lr
                                    break
                            hbranch = (tmp[1], name)
        elif guessmerge and parents[1] == revlog.nullid and hbranch == (-1, ''):
            for key, val in files.iteritems():
                mi = svn.mergeinfo(val, rev.revnum)

                # mergeinfo was not found
                if not mi:
                    continue

                # find the highest value in a list of lists (which are really
                # rev ranges from subversion), e.g. [(34468, 34855, 1), (34899,
                # 37508, 1), (37510, 37512, 1), (41039, 41463, 1)]
                hval = (0, 0, 1)
                hkey = None
                for mpath, mval in mi.iteritems():
                    revs = sorted(mval, key=lambda revrange: revrange[1])[-1:][0]
                    if revs[1] >= hval[1]:
                        hval = revs
                        hkey = mpath

                # now, we look through the mercurial history within the given
                # range for changed file (starting at the highest rev and
                # working our way backwards)
                name = meta.split_branch_path(hkey)[1]
                for r in range(hval[1], hval[0], -1):
                    if (r, name) in meta.revmap:
                        ctx = meta.repo[meta.revmap[(r, name)]]
                        if key in ctx.files():
                            hbranch = (r, name)
                            break

            # still no revision found? try regex!
            if hbranch == (-1, ''):
                regexrevs = util.findrevs(mesg)
                if regexrevs:
                    hgrev = meta.repo.revs('limit(svnrev(%d), 1)' % max(regexrevs))
                    if hgrev:
                        try:
                            parents[1] = meta.repo[int(hgrev[0])].node()
                        except TypeError:
                            # newer mercurial baseset has no indexing
                            parents[1] = meta.repo[int(hgrev.first())].node()

        if (parents[1] == node.nullid
            and hbranch != (-1, '') and hbranch in meta.revmap):
            # at this point, we have no second parent but did find one through
            # the above code
            parents[1] = meta.revmap[hbranch]

        if meta.anchor and parents[0] == revlog.nullid:
            try:
                parents[0] = meta.repo['default'].node()
            except error.RepoLookupError:
                parents[0] = revlog.nullid

        if parents[0] in closedrevs and branch in meta.closebranches:
            continue

        extra = meta.genextra(rev.revnum, branch)
        tag = None
        if branch is not None:
            # New regular tag without modifications, it will be committed by
            # svnmeta.committag(), we can skip the whole branch for now
            tag = meta.get_path_tag(meta.remotename(branch))
            if (tag and tag not in meta.tags
                and branch not in meta.branches
                and branch not in compathacks.branchset(meta.repo)
                and not files):
                continue

        parentctx = meta.repo.changectx(parents[0])
        if tag:
            if parentctx.node() == node.nullid:
                continue
            extra.update({'branch': parentctx.extra().get('branch', None),
                          'close': 1})

        def filectxfn(repo, memctx, path):
            current_file = files[path]
            try:
                data, isexec, islink, copied = current.pop(current_file)
            except IOError:
                return compathacks.filectxfn_deleted_reraise(memctx)
            if isexec is None or islink is None:
                flags = parentctx.flags(path)
                if isexec is None:
                    isexec = 'x' in flags
                if islink is None:
                    islink = 'l' in flags

            if data is not None:
                if islink:
                    if data.startswith('link '):
                        data = data[len('link '):]
                    else:
                        ui.debug('file marked as link, but may contain data: '
                            '%s\n' % current_file)
            else:
                data = parentctx.filectx(path).data()
            return compathacks.makememfilectx(repo,
                                              path=path,
                                              data=data,
                                              islink=islink,
                                              isexec=isexec,
                                              copied=copied)

        meta.mapbranch(extra)

        # if we're anchoring the new branch to the trunk, then look for a
        # previous branch of the same name and put this new commit on top of
        # that one
        if (meta.anchor and 'branch' in extra
            and meta.repo[parents[0]].branch() != extra['branch']
            and extra['branch'] in meta.repo):
            p2 = meta.repo[extra['branch']]
            parents[1] = p2.node()

        current_ctx = context.memctx(meta.repo,
                                     parents,
                                     mesg,
                                     files.keys(),
                                     filectxfn,
                                     meta.authors[rev.author],
                                     date,
                                     extra)

        new_hash = meta.repo.svn_commitctx(current_ctx)
        util.describe_commit(ui, new_hash, branch)
        if (rev.revnum, branch) not in meta.revmap and not tag:
            meta.revmap[rev.revnum, branch] = new_hash
        if tag:
            meta.movetag(tag, new_hash, rev, date)
            meta.addedtags.pop(tag, None)

    # 2. handle branches that need to be committed without any files
    for branch in current.emptybranches:

        if meta.skipbranch(branch):
            continue

        ha = meta.get_parent_revision(rev.revnum, branch)
        if ha == node.nullid:
            continue

        parent_ctx = meta.repo.changectx(ha)
        files = []
        def del_all_files(*args):
            raise IOError(errno.ENOENT, 'deleting all files')

        # True here means nuke all files.  This happens when you
        # replace a branch root with an empty directory
        if current.emptybranches[branch]:
            files = meta.repo[ha].files()

        extra = meta.genextra(rev.revnum, branch)
        meta.mapbranch(extra)

        parents = [ha, node.nullid]
        files = {}

        # if we're anchoring the new branch to the trunk, then look for a
        # previous branch of the same name and put this new commit on top of
        # that one
        if (meta.anchor and 'branch' in extra
            and meta.repo[ha].branch() != extra['branch']
            and extra['branch'] in meta.repo):
            # don't delete util.ignoredfiles since that will mess up tags and
            # externals
            p2 = meta.repo[extra['branch']]
            files = {f:p2[f].data() for f in util.ignoredfiles if f in p2}
            parents[1] = p2.node()

        parent_ctx = meta.repo.changectx(ha)
        def del_all_files(repo, ctx, path):
            if path not in files:
                raise IOError(errno.ENOENT, 'deleting all files')
            return compathacks.makememfilectx(repo,
                                              path=path,
                                              data=files[path],
                                              islink=False,
                                              isexec=False,
                                              copied=False)

        current_ctx = context.memctx(meta.repo,
                                     parents,
                                     mesg,
                                     files.keys(),
                                     del_all_files,
                                     meta.authors[rev.author],
                                     date,
                                     extra)
        new_hash = meta.repo.svn_commitctx(current_ctx)
        util.describe_commit(ui, new_hash, branch)
        if (rev.revnum, branch) not in meta.revmap:
            meta.revmap[rev.revnum, branch] = new_hash

    return closebranches
